(** nat properties *)

(** Chunxiao Lin @ 2006 *)

Require Export EqNat.
Require Export Omega.
Require Export Div2.
Require Export basicutil.

Set Implicit Arguments.
Unset Strict Implicit.


(** ******************************************************************* *)
(** more induction lemmas *)
(** ******************************************************************* *)
Section Ind.
  Variable P : nat -> Prop.
  Variable p0 : P 0.
  Variable p1 : P 1.
  Variable p2 : P 2.
  Variable p3 : P 3.
  Variable p4 : P 4.
  Variable p5 : P 5.
  Variable p6 : P 6.
  Variable p7 : P 7.

  Fixpoint ind2 (n : nat) (pf : forall n, P n -> P (S (S n))) 
    {struct n} : P n :=
      match n as n0 return (P n0) with
	| 0 => p0
	| 1 => p1
	| S (S n0) => pf n0 (ind2 n0 pf) 
      end.

  Fixpoint ind4 (n : nat) (pf : forall n, P n -> P (4 + n))
    {struct n} : P n :=
      match n as n0 return (P n0) with
	| 0 => p0
	| 1 => p1
	| 2 => p2
	| 3 => p3
	| S(S(S(S n0))) => pf n0 (ind4 n0 pf) 
      end.

  Fixpoint ind8 (n : nat) (pf : forall m, P m -> P (8 + m))
    {struct n} : P n :=
      match n as n0 return (P n0) with
	| 0 => p0
	| 1 => p1
	| 2 => p2
	| 3 => p3
	| 4 => p4
	| 5 => p5
	| 6 => p6
	| 7 => p7
	| S(S(S(S( S(S(S(S n0))))))) => pf n0 (ind8 n0 pf) 
      end.
End Ind.


(** ******************************************************************* *)
(** div4 *)
(** ******************************************************************* *)
Definition div4 x := div2 (div2 x).

Lemma div4plus : forall n, 
  div4 (4 + n) = 1 + div4 n.
Proof.
  intros.
  pattern n.
  apply ind4;
    unfold div4; simpl; try omega.
  reflexivity.
Qed.

Lemma div44eq : forall x,
  div4 (4 * x) = x.
Proof.
  induction x.
  reflexivity.
  replace (4 * S x) with (4 + 4 * x).
  rewrite div4plus.
  rewrite IHx.
  reflexivity.
  omega.
Qed.


(** ******************************************************************* *)
(** div8 *)
(** ******************************************************************* *)
Definition div8 x := div2 (div2 (div2 x)).

Lemma div8_plus_8 : forall n, div8 (8 + n) = S (div8 n).
Proof.
  intro n; pattern n; apply ind8; intros;
    simpl; reflexivity.
Qed.

Lemma div8_elim_lt : forall n k, k < div8 n -> n >= 8 * (S k).
Proof.
  intro n; pattern n; apply ind8; intros;
    try (compute in H; omega).
  rewrite div8_plus_8 in H0.
  destruct k.
  omega.
  assert ( k < (div8 m)).
  omega.
  hypreplace H (H _ H1).
  omega.
Qed.

Lemma div8_intro_lt : forall n k, n >= 8 * (S k) -> k < div8 n.
Proof.
  intro n; pattern n; apply ind8; intros; try omega.
  rewrite div8_plus_8.
  destruct k.
  omega.
  assert ( m >= 8 * S k ).
  omega.
  hypreplace H (H _ H1).
  omega.
Qed.


(** ******************************************************************* *)
(** ptr *)
(** ******************************************************************* *)
Section PtrP.
  Fixpoint ptr (n : nat) : Prop :=
    match n with
      | 0 => True
      | S (S (S (S n'))) => ptr n'
      | _ => False
    end.

  Fixpoint atom (n : nat) : Prop :=
    match n with
      | 0 => False
      | 1 => True
      | S (S n') => atom n'
    end.

  Fixpoint bptr (n : nat) : bool :=
    match n with
      | 0 => true
      | S (S (S (S n'))) => bptr n'
      | _ => false
    end.
    
  Fixpoint batom (n : nat) : bool :=
    match n with
      | 0 => false
      | 1 => true
      | S (S n') => batom n'
    end.

  Lemma ptr_dec : forall n, ptr n \/ ~ ptr n.
  Proof.
    intro n.
    pattern n; apply ind4; intros;
      try (left; constructor);
	try (right; intro H; apply H).
    simpl; auto.
  Qed.

  Lemma bptr_dec : forall n, bptr n = true \/ bptr n = false.
  Proof.
    intro n.
    pattern n; apply ind4; intros; simpl;
      intuition.
  Qed.

  Lemma atom_dec : forall n, atom n \/ ~ atom n.
  Proof.
    intro n.
    pattern n; apply ind2; intros;
      try (left; constructor);
	try (right; intro H; apply H).
    simpl; auto.
  Qed.

  Lemma batom_dec : forall n, batom n = true \/ batom n = false.
  Proof.
    intro n.
    pattern n; apply ind2; intros; simpl;
      intuition.
  Qed.

  Lemma bptr_intro : forall n,
    ptr n -> bptr n = true.
  Proof.
    intro n.
    pattern n; apply ind4; intros;
      try reflexivity;
	try (elimtype False; apply H).
    simpl in * ; auto.
  Qed.

  Lemma bptr_elim : forall n,
    bptr n = true -> ptr n.
  Proof.
    intro n.
    pattern n; apply ind4; intros;
      try discriminate;
	try constructor.
    simpl in * ; auto.
  Qed.

  Lemma bptr_intro2 : forall n,
    ~ ptr n -> bptr n = false.
  Proof.
    intro n.
    pattern n; apply ind4; intros;
      try reflexivity;
	try (elimtype False; apply H; constructor).
    simpl in * ; auto.
  Qed.

  Lemma bptr_elim2 : forall n,
    bptr n = false -> ~ ptr n.
  Proof.
    intro n.
    pattern n; apply ind4; red; intros;
      try discriminate;
	try (elimtype False; apply H0; constructor).
    simpl in * ; apply H; auto.
  Qed.

  Lemma batom_intro : forall n,
    atom n -> batom n = true.
  Proof.
    intro n.
    pattern n; apply ind2; intros;
      try reflexivity;
	try (elimtype False; apply H).
    simpl in * ; auto.
  Qed.

  Lemma batom_elim : forall n,
    batom n = true -> atom n.
  Proof.
    intro n.
    pattern n; apply ind2; intros;
      try discriminate;
	try constructor.
    simpl in * ; auto.
  Qed.

  Lemma batom_intro2 : forall n,
    ~ atom n -> batom n = false.
  Proof.
    intro n.
    pattern n; apply ind2; intros;
      try reflexivity;
	try (elimtype False; apply H; constructor).
    simpl in * ; auto.
  Qed.

  Lemma batom_elim2 : forall n,
    batom n = false -> ~ atom n.
  Proof.
    intro n.
    pattern n; apply ind2; red; intros;
      try discriminate;
	try (elimtype False; apply H0; constructor).
    simpl in * ; apply H; auto.
  Qed.

  Lemma atom_not_ptr : forall n,
    atom n -> ~ ptr n.
  Proof.
    intro n.
    pattern n; apply ind4; red; intros;
      try discriminate;
	try (elimtype False; apply H0; constructor).
    simpl in * ; apply H; auto.
    simpl in * ; apply H; auto.
  Qed.

  Lemma ptr_not_atom : forall n,
    ptr n -> ~ atom n.
  Proof.
    intro n.
    pattern n; apply ind4; red; intros;
      try discriminate;
	try (elimtype False; apply H0; constructor).
    simpl in * ; apply H; auto.
    simpl in * ; apply H; auto.
    simpl in * ; apply H; auto.
  Qed.

End PtrP.


(** ******************************************************************* *)
(** beq *)
(** ******************************************************************* *)
Section Beq.
  Lemma beq_true : forall m,
    beq_nat m m = true.
  Proof.
    intro.
    assert (X := beq_nat_refl m).
    symmetry.
    assumption.
  Qed.
  
  Lemma beq_false : forall m n,
    m <> n -> beq_nat m n = false.
  Proof.
    double induction m n; intros;
      try reflexivity.
    destruct H; reflexivity.
    simpl; apply H0.
    intro F; apply H1; auto.
  Qed.

  Lemma beq_dec : forall m n,
    beq_nat m n = true \/ beq_nat m n = false.
  Proof.
    intros.
    destruct (beq_nat m n);
      [left | right]; trivial.
  Qed.
  
End Beq.

Hint Resolve beq_true beq_false : natutil.


(** ******************************************************************* *)
(** blt *)
(** ******************************************************************* *)
Section Blt.

  Fixpoint blt_nat (m n : nat) {struct m} : bool :=
    match m, n with
      | O, S _ => true
      | S m', S n' => blt_nat m' n'
      | _, _ => false
    end.
  
  Lemma blt_true : forall m n,
    m < n ->
    blt_nat m n = true.
  Proof.
    double induction m n; intros.
    elimtype False; omega.
    reflexivity.
    elimtype False; omega.
    simpl.
    apply H0; omega.
  Qed.

  Lemma blt_false : forall m n,
    m >= n -> blt_nat m n = false.
  Proof.
    double induction m n;
      intuition.
    simpl; apply H0. 
    inversion_clear H1; auto with arith.
  Qed.

  Lemma blt_lt : forall m n,
    blt_nat m n = true -> m < n.
  Proof.
    induction m; induction n; intros.
    discriminate H.
    auto with arith.
    discriminate H.
    simpl in H.
    assert (L := IHm _ H).
    auto with arith.
  Qed.

  Lemma blt_ge : forall m n,
    blt_nat m n = false -> m >= n.
  Proof.
    induction m; induction n; intros.
    auto with arith.
    discriminate H.
    auto with arith.
    simpl in H.
    assert (L := IHm _ H).
    auto with arith.
  Qed.

  Lemma blt_dec : forall m n,
    blt_nat m n = true \/ blt_nat m n = false.
  Proof.
    intros.
    destruct (blt_nat m n);
      [left | right]; trivial.
  Qed.

  Lemma beq_not_blt : forall m n,
    beq_nat m n = true -> blt_nat m n = false.
  Proof.
    intros.
    rewrite (beq_nat_eq _ _ (sym_eq H)).
    clear H.
    induction n; simpl.
    reflexivity.
    exact IHn.
  Qed.

  Lemma bgt_not_blt : forall m n,
    blt_nat n m = true -> blt_nat m n = false.
  Proof.
    intros.
    assert (E := blt_lt H).
    apply blt_false.
    auto with arith.
  Qed.

  Lemma bgt_not_beq : forall m n,
    blt_nat n m = true -> beq_nat m n = false.
  Proof.
    intros.
    assert (E := blt_lt H).
    apply beq_false.
    omega.
  Qed.


End Blt.

Hint Resolve blt_true blt_false blt_lt blt_ge 
beq_not_blt bgt_not_blt bgt_not_beq : natutil.


(** ******************************************************************* *)
(** b odd *)
(** ******************************************************************* *)
Section BOdd.

  Fixpoint odd (n : nat) : Prop :=
    match n with
      | O => False
      | S O => True
      | S (S n') => odd n'
    end.
  
  Fixpoint bodd (n : nat) : bool :=
    match n with
      | O => false
      | S O => true
      | S (S n') => bodd n'
    end.

  Lemma bodd_elim1 : forall n,
    bodd n = true -> odd n.
  Proof.
    intro n.
    assert (exists k, 2 * k <= n < 2 * (S k)).
    induction n.
    exists 0; omega.
    destruct IHn.
    compare (2 * x) n; intro e; subst.
    exists x; omega.
    exists (S x); omega.
    destruct H.
    generalize dependent n.
    induction x; intros.
    destruct n.
    inversion H0.
    destruct n.
    simpl; tauto.
    elimtype False; omega.
    assert (exists m, S (S m) = n).
    exists (n - 2); omega.
    destruct H1; subst n.
    simpl in * .
    apply IHx; auto.
    omega.
  Qed.

  Lemma bodd_intro1 : forall n,
    odd n -> bodd n = true.
  Proof.
    intro n.
    assert (exists k, 2 * k <= n < 2 * (S k)).
    induction n.
    exists 0; omega.
    destruct IHn.
    compare (2 * x) n; intro e; subst.
    exists x; omega.
    exists (S x); omega.
    destruct H.
    generalize dependent n.
    induction x; intros.
    destruct n.
    inversion H0.
    destruct n.
    simpl; tauto.
    elimtype False; omega.
    assert (exists m, S (S m) = n).
    exists (n - 2); omega.
    destruct H1; subst n.
    simpl in * .
    apply IHx; auto.
    omega.
  Qed.

  Lemma bodd_intro2 : forall n,
    ~ odd n -> bodd n = false.
  Proof.
    intros.
    assert (~ bodd n = true).
    intro F; apply H; apply (bodd_elim1 F).
    destruct (bodd n); intuition.
  Qed.

  Lemma bodd_elim2 : forall n,
    bodd n = false -> ~ odd n.
  Proof.
    red; intros.
    rewrite (bodd_intro1 H0) in H; discriminate.
  Qed.

End BOdd.


Ltac simplbeq := 
  match goal with
    | [ |- context [(beq_nat ?x ?x)] ] => 
      rewrite beq_true
    | [ H : ?x <> ?y |- context [(beq_nat ?x ?y)] ] => 
      rewrite (beq_false H)
    | [ H : ?y <> ?x |- context [(beq_nat ?x ?y)] ] => 
      rewrite (beq_false (sym_not_eq H))
    | [ |- context [(beq_nat ?x ?y)] ] => 
      rewrite beq_false
  end.


(** ******************************************************************* *)
(** b compare *)
(** ******************************************************************* *)
Definition bcompare_dec : forall m n : nat,
  blt_nat m n = true \/ beq_nat m n = true \/ blt_nat n m = true.
Proof.
  induction m; induction n.
  right; left.
  apply beq_true.
  left.
  apply blt_true; auto with arith.
  right; right.
  apply blt_true; auto with arith.
  simpl.
  apply (IHm n).
Qed.

Ltac simplcmp a b := 
  let RE := fresh "RE" in
    (destruct (bcompare_dec a b) as [RE | [RE | RE]];
      try (rewrite RE); try (rewrite (beq_not_blt RE));
	try (rewrite (bgt_not_blt RE)); try (rewrite (bgt_not_beq RE))).


(** ******************************************************************* *)
(** b test *)
(** ******************************************************************* *)
Section BTest.

  Inductive btype : Set :=
    | bteq 
    | btne
    | btlt
    | btge.

  Definition btest (bt : btype) (x y : nat) : bool :=
    match bt with
      | bteq => beq_nat x y
      | btne => if (beq_nat x y) then false else true
      | btlt => blt_nat x y
      | btge => if (blt_nat x y) then false else true
    end.
  
  Definition btest_prop (bt : btype) (x y : nat) : Prop :=
    match bt with
      | bteq => x = y
      | btne => x <> y
      | btlt => x < y
      | btge => x >= y
    end.

  Lemma eq_nat_dec : forall x y : nat,
    x = y \/ x <> y.
  Proof.
    induction x; induction y.
    left; reflexivity.
    destruct IHy.
    rewrite <- H; right; auto.
    right; auto.
    right; auto.
    destruct (IHx y); [left | right]; auto.
  Qed.

  Lemma lt_nat_dec : forall x y : nat,
    x < y \/ x >= y.
  Proof.
    induction x; induction y.
    right; auto.
    left; auto with arith.
    right; auto with arith.
    destruct (IHx y); [left | right]; auto with arith.
  Qed.

  Lemma btest_prop_dec : forall bt x y,
    btest_prop bt x y \/ ~ btest_prop bt x y.
  Proof.
    intros.
    destruct bt; simpl;
      destruct (eq_nat_dec x y);
	destruct (lt_nat_dec x y);
	  intuition.
  Qed.

  Lemma btest_prop_true : forall bt x y,
    btest_prop bt x y -> btest bt x y = true.
  Proof.
    destruct bt; simpl.
    intros; subst x; apply beq_true.
    intros; simplbeq.
    reflexivity.
    intros; apply (blt_true H).
    intros; rewrite blt_false; trivial.
  Qed.

  Lemma btest_prop_false : forall bt x y,
    ~ btest_prop bt x y -> btest bt x y = false.
  Proof.
    destruct bt; simpl.
    intros; apply (beq_false H).
    intros; replace x with y; 
      [simplbeq; reflexivity | intuition].
    intros; apply blt_false.
    omega.
    intros; rewrite blt_true; [reflexivity | omega].
  Qed.

End BTest.
  
Ltac simplif :=
  match goal with
    | [ |- context [(if beq_nat ?x ?y then _ else _)]] =>
      let R := fresh "R" in
        (destruct (eq_nat_dec x y) as [R | R]; [rewrite R | idtac]; simplbeq)
    | [ |- context [(if blt_nat ?x ?y then _ else _)]] =>
      let R := fresh "R" in
        (destruct (lt_nat_dec x y) as [R | R]; 
          [rewrite (blt_true R) | rewrite (blt_false R)])
  end.


(** ******************************************************************* *)
(** mod8 *)
(** ******************************************************************* *)
Fixpoint mod8 (n : nat) : nat :=
  match n with
    | S(S(S(S (S(S(S(S n'))))))) => mod8 n'
    | n' => n'
  end.

Lemma mod8_intro : forall n k, n = 8 * k -> mod8 n = 0.
Proof.
  intro n.
  pattern n; apply ind8; intros; simpl;
    try omega.
  induction k.
  omega.
  apply H with k.
  omega.
Qed.

Lemma mod8_elim : forall n, mod8 n = 0 -> exists k, n = 8 * k.
Proof.
  intro n.
  pattern n; apply ind8; intros; 
    try (simpl in H; elimtype False; omega).
  exists 0; reflexivity.
  destruct (H H0).
  exists (S x).
  omega.
Qed.

Lemma mod8_mul_8 : forall n, mod8 (8 * n) = 0.
Proof.
  intros.
  apply mod8_intro with n.
  reflexivity.
Qed.

Lemma mod8_plus4 : forall l, mod8 l = 0 -> mod8 (l + 4) = 4.
Proof.
  intro l; pattern l.
  apply ind8; simpl; intros; try discriminate; firstorder.
Qed.

Lemma mod8_ptr : forall n, mod8 n = 0 -> ptr n.
Proof.
  intro n; pattern n; apply ind8;
    simpl; intros; try discriminate; auto.
Qed.

Lemma mod8_plus4_ptr : forall n, mod8 n = 0 -> ptr (n + 4).
Proof.
  intro n; pattern n; apply ind8;
    simpl; intros; try discriminate; auto.
Qed.

