(** Basic machine model *)

(** register file *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Import List.


(** ******************************************************************* *)
(** registers *)
(** ******************************************************************* *)
Section Reg.

  Inductive reg : Set :=
    | r0 | r1 | r2 | r3 | r4 | r5 | r6 | r7 
    | r8 | r9 | r10 | r11 | r12 | r13 | r14 | r15 
    | r16 | r17 | r18 | r19 | r20 | r21 | r22 | r23 
    | r24 | r25 | r26 | r27 | r28 | r29 | r30 | r31.
  
  Definition beq_reg rx ry : bool :=
    match rx, ry with
      | r0, r0 => true
      | r1, r1 => true
      | r2, r2 => true
      | r3, r3 => true
      | r4, r4 => true
      | r5, r5 => true
      | r6, r6 => true
      | r7, r7 => true
      | r8, r8 => true
      | r9, r9 => true
      | r10, r10 => true
      | r11, r11 => true
      | r12, r12 => true
      | r13, r13 => true
      | r14, r14 => true
      | r15, r15 => true
      | r16, r16 => true
      | r17, r17 => true
      | r18, r18 => true
      | r19, r19 => true
      | r20, r20 => true
      | r21, r21 => true
      | r22, r22 => true
      | r23, r23 => true 
      | r24, r24 => true
      | r25, r25 => true
      | r26, r26 => true
      | r27, r27 => true
      | r28, r28 => true
      | r29, r29 => true
      | r30, r30 => true
      | r31, r31 => true
      | _, _ => false
    end.
  
  Lemma eq_reg_dec : forall (r r' : reg),
    r = r' \/ r <> r'.
  Proof.
    intros.
    compare r r'; intro e; subst;
      [left | right]; trivial.
  Qed.
    
  Ltac  comparer r r' := destruct(eq_reg_dec r r') as [e | e]; subst.

  Lemma beq_reg_intro1 : forall r,
    beq_reg r r = true.
  Proof.
    intros.
    case r; reflexivity.
  Qed.
  
  Lemma beq_reg_intro2 : forall r r',
    r <> r' -> beq_reg r r' = false.
  Proof.
    intros.
    destruct r; destruct r'; 
      try reflexivity;
      elimtype False; apply H;
	reflexivity.
  Qed.
  
  Lemma beq_reg_dec : forall r r',
    beq_reg r r' = true \/ beq_reg r r' = false.
  Proof.
    intros.
    comparer r r'.
    left; apply beq_reg_intro1.
    right; apply beq_reg_intro2; trivial.
  Qed.

  Lemma beq_reg_elim1 : forall r r', beq_reg r r' = true -> r = r'.
  Proof.
    intros.
    destruct r; destruct r'; 
      simpl; intros; 
	try discriminate; reflexivity.
  Qed.

  Lemma beq_reg_elim2 : forall r r', beq_reg r r' = false -> r <> r'.
  Proof.
    red; intros.
    subst r'; rewrite (beq_reg_intro1 r) in H.
    discriminate.
  Qed.

End Reg.

Ltac  comparer r r' := destruct(eq_reg_dec r r') as [e | e]; subst.

Require Import basicutil.

Ltac solvereg :=
  match goal with 
    | [Hbeq : beq_reg ?r1 ?r2 = true |- _ ] =>
      hypreplace Hbeq (beq_reg_elim1 Hbeq); subst; solvereg
    | [Hbeq : beq_reg ?r1 ?r2 = false |- _ ] =>
      hypreplace Hbeq (beq_reg_elim1 Hbeq); solvereg
    | [ |- context [(beq_reg ?x ?x)] ] => 
      rewrite beq_reg_intro1; solvereg
    | [ H : ?x <> ?y |- context [(beq_reg ?x ?y)] ] =>
      rewrite (beq_reg_intro2 H); solvereg
    | [ H : ?y <> ?x |- context [(beq_reg ?x ?y)] ] =>
      rewrite beq_reg_intro2; [solvereg | auto]
    | _ => idtac
  end.

Section RFile.

  (** rfile is a total map from reg name to nat *) 
  Definition rfile := reg -> nat.

  Definition rfile_get (rf : rfile) (r : reg) := 
    if beq_reg r0 r then 0 else (rf r).
  
  Definition rfile_set rf r val : rfile := 
    fun r' => if beq_reg r' r then val else rfile_get rf r'.

  Lemma rfile_set_get_eq : forall rf r val,
    r <> r0 ->
    rfile_get (rfile_set rf r val) r = val.
  Proof.
    intros.
    unfold rfile_set; unfold rfile_get.
    solvereg.
    reflexivity.
  Qed.

  Lemma rfile_set_get_neq : forall rf r r' val,
    r <> r' ->
    rfile_get (rfile_set rf r' val) r = rfile_get rf r.
  Proof.
    intros.
    unfold rfile_set; unfold rfile_get.
    comparer r0 r; solvereg; reflexivity.
  Qed.

End RFile.

(** ******************************************************************* *)
(** MIPS register conventions *)
(** ******************************************************************* *)
Notation ra := r31.  (* return address proc *)
Notation zero := r0. (*always zero *)
Notation ast := r1.  (* assembler temporary *)
Notation v0 := r2.   (*v0 - v1 return values *)
Notation v1 := r3.
Notation a0 := r4.   (*a0 - a3 arguments *)
Notation a1 := r5.
Notation a2 := r6.
Notation a3 := r7.
Notation t0 := r8.   (*t0 - t9 caller saved temporary *)
Notation t1 := r9.
Notation t2 := r10. 
Notation t3 := r11.
Notation t4 := r12.
Notation t5 := r13.
Notation t6 := r14.
Notation t7 := r15.
Notation t8 := r24.
Notation t9 := r25.
Notation s0 := r16.  (*s0 - s7 callee saved temporary *)
Notation s1 := r17.
Notation s2 := r18.
Notation s3 := r19.
Notation s4 := r20.
Notation s5 := r21.
Notation s6 := r22.
Notation s7 := r23.
Notation k0 := r26.
Notation k1 := r27.
Notation gp := r28.
Notation sp := r29.
Notation fp := r30.

Ltac substb :=
  match goal with 
    | [Hbeq : beq_reg ?r1 ?r2 = true |- _ ] =>
      hypreplace Hbeq (beq_reg_elim1 Hbeq); subst
  end.

Ltac simplrf :=
  repeat (rewrite rfile_set_get_neq; [idtac | discriminate]);
  repeat rewrite rfile_set_get_eq;
  repeat (rewrite rfile_set_get_neq; [idtac | discriminate]);
  repeat rewrite rfile_set_get_eq.


(** ******************************************************************* *)
(** register list *)
(** ******************************************************************* *)
Notation rlst_in := (fun r rl => In r rl).

Fixpoint rlst_nin (r : reg) (rl : list reg) {struct rl} : Prop :=
  match rl with
    | r' :: rl' => r <> r' /\ rlst_nin r rl'
    | _ => True
  end.

Definition rlst_subset (rr1 rr2 : list reg) : Prop :=
  forall r, rlst_in r rr1 -> rlst_in r rr2.
