(** GC Safety for TAL *)

(** Type system side interface *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export tal.
Require Export reach.


(** ******************************************************************* *)
(** reachability from TAL regs *)
(** ******************************************************************* *)
Notation rchtal := (fun s x => (reachs (hp_of s) (map (rget s) tal_regs') x)).

Notation rchtalb := (fun s x => (reachsb (hp_of s) (map (rget s ) tal_regs') x )).

Lemma rchtal_tal_reg' : forall s r,
  vptr (rget s r) -> rlst_in r tal_regs' -> rchtalb s (rget s r) = true.
Proof.
  intros.
  apply reachsb_intro with (hp := hp_of s) 
    (s := map (rget s) tal_regs') (x := rget s r).
  destruct H0 as [Z | [Z | [Z | Z]]]; try (exists (rget s r)); 
    subst; firstorder; try (constructor; firstorder).
Qed.


(** ******************************************************************* *)
(** a Phi only talks about the reachable part of the heap *)
(** ******************************************************************* *)
Notation dhspec_prune := (fun Phi st => 
  (fun l' => if rchtalb st l' then Phi l' else None)).


(** ******************************************************************* *)
(** heap prune lemmas *)
(** ******************************************************************* *)
Lemma heap_prune_WFdhspec : forall Phi st,
  WFdhspec Phi ->
  WFdhspec (dhspec_prune Phi st).
Proof.
  red; intros.
  destruct (reachsb (hp_of st) (map (rget st) tal_regs') l).
  apply (H _ _ _ _ _ H0).
  discriminate.
Qed.

Lemma heap_prune_not_vptr_WFw : forall T Psi Phi s v,
  WFdhspec Phi ->
  (~ vptr v) ->
  WFw Psi Phi v T ->
  WFw Psi (dhspec_prune Phi s) v T.
Proof.
  intros T Psi Phi s v Hphi Hv Hw.
  induction Hw; rename Phi into XXX; 
    rename Psi into Phi; rename XXX into Psi.

  constructor.
  constructor.
  apply WFw_code; trivial.

  destruct C1; destruct C2;
    destruct (Hphi _ _ _ _ _ H).
  tauto.
  
  constructor.
  apply IHHw; trivial.

  apply WFw_sub with T; trivial.
  apply IHHw; trivial.
Qed.

Lemma heap_prune_rchtal_WFw : forall T Psi Phi s v,
  WFdhspec Phi ->
  rchtal s v ->
  WFw Psi Phi v T ->
  WFw Psi (dhspec_prune Phi s) v T.
Proof.
  intros T Psi Phi s v Hphi Hr Hw.
  induction Hw; rename Phi into XXX; 
    rename Psi into Phi; rename XXX into Psi.

  constructor.
  constructor.
  apply WFw_code; trivial.

  apply WFw_mem with C1 C2; trivial.
  rewrite reachsb_intro; trivial.
  
  constructor.
  apply IHHw; trivial.

  apply WFw_sub with T; trivial.
  apply IHHw; trivial.
Qed.

Lemma heap_prune_reg_WFw : forall T Psi Phi s r,
  WFdhspec Phi ->
  rlst_in r tal_regs' ->
  WFw Psi Phi (rget s r) T ->
  WFw Psi (dhspec_prune Phi s) (rget s r) T.
Proof.
  intros T Psi Phi s r Hphi Hr Hw.
  assert (Hv : exists v, rget s r = v).
  exists (rget s r); reflexivity.
  destruct Hv as [v Hv].
  rewrite Hv in Hw.
  induction Hw; rewrite Hv;
    rename Phi into XXX; rename Psi into Phi; rename XXX into Psi.
  
  constructor.
  constructor.
  apply WFw_code; trivial.

  subst addr.
  apply WFw_mem with C1 C2; trivial.
  rewrite rchtal_tal_reg'; trivial.
  destruct C1; destruct C2;
    destruct (Hphi _ _ _ _ _ H); assumption.
  
  constructor.
  subst w.
  apply IHHw; trivial.

  subst w.
  apply WFw_sub with T; trivial.
  apply IHHw; trivial.
Qed.

Lemma heap_prune_rchtal_WFcell : forall Psi Phi s v b T,
  WFdhspec Phi ->
  rchtal s v ->
  WFcell Psi Phi v (b, T) ->
  WFcell Psi (dhspec_prune Phi s) v (b, T).
Proof.
  intros.
  induction b.
  inversion H1; subst.
  constructor.
  apply heap_prune_rchtal_WFw; trivial.
  constructor.
Qed.

Lemma heap_prune_not_vptr_WFcell : forall Psi Phi s v b T,
  WFdhspec Phi ->
  (~ vptr v) ->
  WFcell Psi Phi v (b, T) ->
  WFcell Psi (dhspec_prune Phi s) v (b, T).
Proof.
  intros.
  induction b.
  inversion H1; subst.
  constructor.
  apply heap_prune_not_vptr_WFw; trivial.
  constructor.
Qed.

Lemma heap_prune_WFrf : forall G s Psi Phi, 
  WFdhspec Phi ->
  WFrf Psi Phi (rf_of s) G ->
  WFrf Psi (dhspec_prune Phi s) (rf_of s) G.
Proof.
  intros G s Psi Phi Hphi [Hr Hwf].
  split; trivial.
  intros.
  hypreplace Hr (Hr _ _ H).
  comparer r zero.
  destruct Hwf as [_ [_ [_ Hz]]].
  destruct Hz as [Hz | [ Hz | Hz]];
    rewrite Hz in H; try discriminate;
      injection H; intros; subst; constructor.
  assert (F : WFw Psi (dhspec_prune Phi s) (rget s r) T);
    [apply heap_prune_reg_WFw; trivial | apply F].
  destruct Hwf as [Hb [Hd [Hc Hz]]].
  apply tal_regs'_intro; trivial.
  apply Hd.
  apply regty_indom_intro with T; trivial.
Qed.

Lemma heap_prune_WFdh : forall s Psi Phi,
  WFdh Psi Phi (hp_of s) Phi ->
  WFdh Psi (dhspec_prune Phi s) (hp_of s) (dhspec_prune Phi s).
Proof.
  intros s Psi Phi [Hhp Hwf].
  split.

  intros l' C1 C2.
  destruct (reachsb_dec (hp_of s) (map (rget s) tal_regs') l')
    as [R | R]; rewrite R; intros.
  destruct (Hhp _ _ _ H) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].
  exists v1; exists v2.
  repeat (split; trivial).

  destruct C1.
  destruct (vptr_dec v1) as [V | V].
  apply heap_prune_rchtal_WFcell; trivial.
  destruct (reachsb_elim R) as [a [Hin Hrch]].
  exists a; split; trivial.
  apply rch_next2 with l'; trivial.
  left; assumption.
  apply heap_prune_not_vptr_WFcell; trivial.

  destruct C2.
  destruct (vptr_dec v2) as [V | V].
  apply heap_prune_rchtal_WFcell; trivial.
  destruct (reachsb_elim R) as [a [Hin Hrch]].
  exists a; split; trivial.
  apply rch_next2 with l'; trivial.
  right; assumption.
  apply heap_prune_not_vptr_WFcell; trivial.
  
  discriminate.

  apply heap_prune_WFdhspec; assumption.
Qed.


(** ******************************************************************* *)
(** what we want from gc *)
(** ******************************************************************* *)
Definition gc_step s s' :=
  (forall x, rchtal s x -> 
    (hget s x = hget s' x /\ hget s (x + WSIZE) = hget s' (x + WSIZE))) /\
  (forall r, rlst_in r tal_regs' -> rget s r = rget s' r).


(** ******************************************************************* *)
(** gc_step lemmas *)
(** ******************************************************************* *)
Lemma gc_step_rchtal : forall s s' x,
  gc_step s s' -> rchtal s x -> rchtal s' x.
Proof.
  intros s s' x [Hh Hr] [a [Hin Hrch]].
  exists a; split.
  simpl.
  destruct Hin as [e | [e | [e | e]]]; try (subst a).
  rewrite (Hr s1); intuition.
  rewrite (Hr s2); intuition.
  rewrite (Hr ra); intuition.
  tauto.
  
  apply reach_heap_eqv_reach with (hp_of s); trivial.
  unfold hget in Hh.
  intros; apply Hh.
  exists a; intuition.
Qed.

Lemma gc_step_WFdh : forall s s' Psi Phi,
  gc_step s s' ->
  WFdh Psi (dhspec_prune Phi s) (hp_of s) (dhspec_prune Phi s) ->
  WFdh Psi (dhspec_prune Phi s) (hp_of s') (dhspec_prune Phi s).
Proof.
  intros s s' Psi Phi [Hh Hr] [Hhp Hwf].
  split; trivial.

  intros l' C1 C2 Hc.
  destruct (Hhp _ _ _ Hc) as [v1 [v2 [Hv1 [Hv2 [Hvc1 Hvc2]]]]].
  exists v1; exists v2.
  destruct (reachsb_dec (hp_of s) (map (rget s) tal_regs') l') as [R | R].
  unfold hget in Hh; destruct (Hh _ (reachsb_elim R)).
  repeat (split; trivial).
  rewrite <- H; trivial.
  rewrite <- H0; trivial.
  rewrite R in Hc; discriminate.
Qed.


(** ******************************************************************* *)
(** gc_step guarantees WFst *)
(** ******************************************************************* *)
Theorem gc_step_WFst : forall s s' Psi G,
  gc_step s s' -> WFst Psi G s -> WFst Psi G s'.
Proof.
  intros s s' Psi G Hgc [Phi [Hdh Hrf]].
  exists (dhspec_prune Phi s); split.
  apply gc_step_WFdh; trivial.
  apply heap_prune_WFdh; trivial.
  destruct Hdh as [_ Hwfdh].
  destruct (heap_prune_WFrf Hwfdh Hrf) as [Hrf' Hwf'].
  split; trivial.
  intros.
  destruct Hwf' as [_ [Hin [_ Hwf']]].
  destruct Hgc as [_ Hr].
  comparer r zero.
  destruct Hwf' as [Z | [ Z | Z]]; rewrite H in Z; try discriminate;
    injection Z; intros; subst; constructor.
  unfold rget in Hr; rewrite <- Hr.
  intuition.
  apply tal_regs'_intro; trivial.
  apply Hin.
  apply regty_indom_intro with T; assumption.
Qed.



(* Lemma gc_step_sym : forall s s', *)
(*   gc_step s s' -> gc_step s' s. *)
(* Proof. *)
(*   unfold gc_step; intros s s' [Hh Hr]. *)
(*   intuition. *)
(*   split. *)
(*   intros x [a [Hin Hrch]]. *)
(*   destruct (Hh x). *)
(*   exists a. *)
(*   split. *)
(*   destruct Hin as [e |[e |[e | e]]]; subst; simpl. *)
(*   rewrite (Hr s1); intuition. *)
(*   rewrite (Hr s2); intuition. *)
(*   rewrite (Hr ra); intuition. *)
(*   tauto. *)

(*   apply reach_heap_eqv_reach with (hp_of s'). *)
(*   destruct H  *)

(*   rewrite <- (H1 _ H). reflexivity. *)
(* Qed. *)

(* here are all the info TAL knows about gc: *)
(* 1. the reachable part of the two states are equal *)
(* 2. the tal regs are equal *)
(* Definition gc_step s s' := *)
(*   (forall x, rchtal s x ->  *)
(*     (hget s x = hget s' x \/ hget s (x + WSIZE) = hget s (x + WSIZE))) /\ *)
(*   (forall r, rlst_in r tal_regs' -> rget s r = rget s' r). *)

(* we will have to prove the following: *)
(* gc_step s s' -> WFst s -> WFst s' *)

(* Theorem tal_gc_safety : forall s s' Psi G, *)
(*   gc_step s s' -> WFst Psi G s -> WFst Psi G s'. *)
(* Proof. *)

(* remove an unreachable object from heap, WFst still holds *)


(* This is a wrong way *)

(* for the gc preseved heaps, there exists an minimal Phi, *)
(* only talks about the reachable part of the heap. *)

(* dom(Phi') \subseteq (rchtal s addr) *)
(* for this part, the two dhspec are equal *)
(* Phi' is a minimal dhspec regrading to Phi *)
(* Inductive miniPhi : state -> dhspec -> dhspec -> Prop := *)
(*   | miniPhi_rch :  *)
(*     forall s Phi Phi', *)
(*       (forall addr x,  *)
(* 	Phi' addr = Some x  <-> *)
(* 	((rchtal s addr \/ rchtal s (addr - 4)) /\ Phi addr = Some x)) -> *)
(*       miniPhi s Phi Phi'. *)


(* Lemma gc_safety_WFrf : forall Psi Phi Phi' G s, *)
(*   miniPhi s Phi Phi' -> *)
(*   WFrf Psi Phi (rf_of s) G -> *)
(*   WFrf Psi Phi' (rf_of s) G. *)
(* Proof. *)
(*   intros Psi Phi Phi' G s Hm [Hrf Hwf]. *)
(*   split; trivial. *)
(*   intros r T Hr. *)
(*   hypreplace Hrf (Hrf _ _ Hr). *)
(*   assert (exists v, v = (rfile_get (rf_of s) r)). *)
(*   exists (rfile_get (rf_of s) r); reflexivity. *)
(*   destruct H as [v Hv]. *)
(*   rewrite <- Hv in Hrf. *)
(*   rewrite <- Hv. *)
(*   induction Hrf; rename Psi into XXX; rename Phi into Psi; rename XXX into Phi. *)
(*   skipgoal. *)
(*   skipgoal. *)
(*   skipgoal. *)
(*   skipgoal. *)
(* Lemma gc_ *)
(*   apply  *)
(*   constructor. *)
(*   constructor. *)
(*   apply WFw_code with Gp; trivial. *)
(*   induction Hm. *)
(*   destruct Hwf as [Hb [Hs [Hc Hz]]]. *)
(*   assert (Hz' : r <> zero). *)
(*   intro F; subst; destruct Hz as [Z | Z]; rewrite Z in Hr; discriminate. *)
(*   assert (Hin := tal_regs'_intro (Hs _ (regty_indom_intro Hr)) Hz'). *)
(*   apply WFw_mem with C1 C2; trivial. *)
  
(*   subst addr. *)
(*   destruct (H3 (rfile_get (rf_of s) r) C1) as [_ Hc1]; clear H3. *)
(*   apply Hc1; split; trivial. *)
(*   left. *)
(*   exists (rget s r); split. *)
(*   destruct Hin as [Z | [Z | [Z | Z]]]; subst; simpl; tauto. *)
(*   apply rch_refl. *)

(*   subst addr. *)
(*   destruct (H3 ((rfile_get (rf_of s) r) + 4) C2) as [_ Hc2]; clear H3. *)
(*   apply Hc2; split; trivial. *)
(*   right. *)
(*   simplpm. *)
(*   exists (rget s r); split. *)
(*   destruct Hin as [Z | [Z | [Z | Z]]]; subst; simpl; tauto. *)
(*   apply rch_refl. *)

(*   constructor. *)
(*   apply IHHrf; trivial. *)
