(** SCAP on LOCAP. *)

(** Simpl VCGen. *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export scap2.
Require Export seplog.


(** ******************************************************************* *)
(** VCGen *)
(** ******************************************************************* *)
Notation no_pre := (fun s => True).

Notation id_guar := (fun s => fun s' => s = s').

Notation ret_spec := (pairT no_pre id_guar).

Fixpoint vc (P : chspec) (is : iseq) {struct is} : optionT (prodT spred guar)  :=
  match is with
    | seq c is' =>
      match vc P is' with
	| SomeT (pairT p' g') =>
	  SomeT (pairT 
	    (fun s => opt_pred (cstep c s) (fun s' => p' s'))
	    (fun s => fun s'' => opt_pred (cstep c s) (fun s' => g' s' s'')))
	| NoneT => NoneT (prodT spred guar)
      end
    | jr r31 => SomeT ret_spec
    | j target => 
      match (chspec_get2 P target) with
	| SomeT bsp => SomeT bsp
	| NoneT => NoneT (prodT spred guar)
      end
    | jal fret target =>
      match (chspec_get2 P target), (chspec_get2 P fret) with
	| SomeT (pairT p' g'), SomeT (pairT p'' g'') =>
	  SomeT (pairT 
	  (fun s => p' (rset s r31 fret) /\
	    (forall s', g' (rset s r31 fret) s' -> p'' s'))
	  (fun s => fun s'' => exists s', g' (rset s r31 fret) s' /\
	    g'' s' s''))
	| _, _ => NoneT (prodT spred guar)
      end
    | br bt rs rt offset is' =>
      match vc P is' with
	| SomeT (pairT p' g') =>
	  SomeT (pairT 
	    (fun s => 
	      if btest bt (rget s rs) (rget s rt)
		then 
		  (opt_predT (chspec_get2 P offset) (fun bsp =>
		let (p', _) := bsp in (p' s)))
		else (p' s))
	    (fun s => fun s' =>
	      if btest bt (rget s rs) (rget s rt) 
		then
		  (opt_predT (chspec_get2 P offset) (fun bsp =>
		let (_, g') := bsp in (g' s s')))
		else (g' s s')))
	| NoneT => NoneT (prodT spred guar)
      end
    | _ => NoneT (prodT spred guar)
  end.

Fixpoint ra_eq_pred (is : iseq) (P : chspec) {struct is} : Prop :=
  match is with
    | seq c is' => ra_eq_pred is' P
    | br bt rs rt offset is' => ra_eq_pred is' P
    | jal fret target => 
      match (chspec_get2 P target) with
	| SomeT (pairT p g) => 
	  forall s s', g s s' -> rget s ra = rget s' ra
	| NoneT => True
      end
    | _ => True
  end.

Theorem vc_sound : forall is P p g,
  (vc P is) = SomeT (pairT p g) ->
  ra_eq_pred is P ->
  scap_iseq_ok P p g is.
Proof.
  induction is; simpl; intros.

  hypreplace IHis (IHis P).
  destruct (vc P is) as [[p' g'] | ]; [idtac | discriminate H].
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  split; intros.
  destruct (cstep c s) as [s' | ]; [idtac | elimtype False; exact H].
  tauto.
  auto.

  hypreplace IHis (IHis P).
  destruct (vc P is) as [[p'' g''] | ]; [idtac | discriminate H].
  injection H; intros; subst p; subst g; clear H.
  exists p''; exists g''.
  split; intros.
  auto.
  destruct (btest b (rget s r) (rget s r0)).
  destruct (chspec_get2_dec P n) as [ E | [[p' g'] E]]; 
    rewrite E; rewrite E in H.
  destruct H.
  exists p'; exists g'.
  repeat (split; trivial).
  apply (chspec_get2_SomeT_in E).
  tauto.

  destruct (chspec_get2_dec P n) as [ E | [[p' g'] E]]; 
    rewrite E in H.
  discriminate.
  injection H; intros; subst; clear H.
  exists p; exists g.
  split.
  apply (chspec_get2_SomeT_in E).
  tauto.

  destruct (chspec_get2_dec P n0) as [ E | [[p' g'] E]]; 
    rewrite E in H.
  discriminate.
  destruct (chspec_get2_dec P n) as [ E2 | [[p'' g''] E2]]; 
    rewrite E2 in H.
  discriminate.
  injection H; intros; subst; clear H.
  exists p'; exists g'; exists p''; exists g''.
  intuition.
  apply (chspec_get2_SomeT_in E).
  apply (chspec_get2_SomeT_in E2).
  exists s'; intuition.
  rewrite E in H0; auto.

  destruct r; try (discriminate H).
  injection H; intros; subst p; subst g; tauto.
Qed.

Notation "s |||- A" := (hp_of s ||- A) (at level 80).


(** ******************************************************************* *)
(** tactics *)
(** ******************************************************************* *)
Ltac simplpg :=
  match goal with
    | [ |- exists p, exists g, 
      SomeT (pairT ?p0 ?g0) = SomeT (pairT p g)] =>
    exists p0; exists g0; reflexivity
  end.

Ltac simplvc :=
  match goal with
    | [ |- scap_iseq_ok ?Psi ?pre ?guar ?is ] =>
      let E := fresh "E" in let p := fresh "p" in let g := fresh "g" in
	let E1 := fresh "E" in let E2 := fresh "E" in
	  (assert (E : exists p, exists g, vc Psi is = SomeT (pairT p g));
	    [simpl; simplpg | destruct E as [p [g E]]];
	    apply scap_iseq_ok_weakening with (p' := p) (g' := g);
	      [injection E; intros E1 E2; subst p; subst g; clear E |
		apply vc_sound; simpl; auto])
  end.

Ltac simpllw x n :=
  match goal with
    | [ |- context [(hget ?s ?addr)]] =>
      let Hp := fresh "Hp" in
	(assert (Hp : hget s addr = Some x);
	  [ unfold hget; apply cont_get_eq1 with (A := top);
	    simplsep 2 n | rewrite Hp; clear Hp; simplst])
    | [ |- context [(heap_get ?h ?addr)]] =>
      let Hp := fresh "Hp" in
	(assert (Hp : heap_get h addr = Some x);
	  [ apply cont_get_eq1 with (A := top);
	    simplsep 2 n | rewrite Hp; clear Hp; simplst])
    | [ |- context [(?h ?addr)]] =>
      let Hp := fresh "Hp" in
	(assert (Hp : heap_get h addr = Some x);
	  [ apply cont_get_eq1 with (A := top);
	    simplsep 2 n | rewrite Hp; clear Hp; simplst])
  end.

Ltac simplsw' sg v:=
  match goal with
    | |- (heap_set ?h ?addr ?val) ||- ?addr |-> ?val => 
      apply cont_update with (v1 := v)
    | |- (heap_set ?h ?addr ?val) ||- sconj (?addr |-> ?val) ?A  => (* apply update *)
      apply cont_update1 with (A := A) (v1 := v)
	
    | [NI : ~ set_in ?x ?s |-
      (heap_set ?h ?addr ?val) ||- fas ?P (set_add ?x ?s)] => 
    apply fas_intro1; [assocr; simplsw' (S sg) v | assumption]
      
    | [NI : ~ set_in ?x ?s |-
      (heap_set ?h ?addr ?val) ||- sconj (fas ?P (set_add ?x ?s)) _] => 
    apply fas_intro; [assocr; simplsw' (S sg) v | assumption]
      
    | |- (heap_set ?h ?addr ?val) ||- sconj _ _  => (* apply update *)
      match eval compute in sg with
	| O => idtac
	| S _ => assocr; apply sconj_sym; assocr; simplsw' (pred sg) v
      end
  end.

Ltac simplsw sg sh v :=
  assocr; simplsw' sg v; simplsep sg sh.

Ltac lwomega :=
  match goal with
    | [H : _ ||- _ |- _] =>
      clear H; lwomega
    | _ => abstract omega
  end.

Lemma call_chain : forall s0 s (p p':spred) (g g' g'':guar),
  (p s /\ forall s', g s s' ->
    p' s' /\ forall s'', g' s' s'' ->
      g'' s0 s'') ->
  (p s /\ forall s', g s s' -> p' s') /\
  (forall s'', (exists s', g s s' /\ g' s' s'') ->
    g'' s0 s'').
Proof.
  intuition.
  destruct (H1 _ H); trivial.
  destruct H as [s' [A B]].
  destruct (H1 _ A) as [C D].
  auto.
Qed.
